# -*- coding: utf-8 -*-
# pylint: disable=r0904, c0415, c0413, r0913, w0212
""" unittests for dkb_robo """
import sys
import os
from datetime import date
import unittest
import logging
import json
from unittest.mock import patch, Mock, MagicMock, mock_open
from bs4 import BeautifulSoup
from mechanicalsoup import LinkNotFoundError
import io
sys.path.insert(0, '.')
sys.path.insert(0, '..')
from dkb_robo.portfolio import ProductGroup, AccountItem, CardItem, DepotItem, Overview, Amount, CardItem, Person, DepotItem

def json_load(fname):
    """ simple json load """

    with open(fname, 'r', encoding='utf8') as myfile:
        data_dic = json.load(myfile)

    return data_dic


class TestProductGroup(unittest.TestCase):
    """ test class """

    def setUp(self):
        """ setup """
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        self.productgroup = ProductGroup()

    def test_001__uid2names(self):
        """ test uid2names one item"""
        data_ele = {'attributes': {'productSettings': {'product': {'uid': {'name': 'name'}}}}}
        self.assertEqual({'uid': 'name'}, self.productgroup._uid2names(data_ele))

    def test_002__uid2names(self):
        """ test uid2names two items """
        data_ele = {'attributes': {'productSettings': {'product1': {'uid1': {'name': 'name1'}}, 'product2': {'uid2': {'name': 'name2'}}}}}
        self.assertEqual({'uid1': 'name1', 'uid2': 'name2'}, self.productgroup._uid2names(data_ele))

    def test_003__uid2names(self):
        """ test uid2names malformed input """
        data_ele = {'attributes': {'foo': 'bar'}}
        self.assertFalse({}, self.productgroup._uid2names(data_ele))

    def test_004__uid2names(self):
        """ test uid2names malformed input """
        data_ele = {'attributes': {'productSettings': {'foo': 'bar'}}}
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertFalse({}, self.productgroup._uid2names(data_ele))
        self.assertIn('WARNING:dkb_robo.portfolio:uid2name mapping failed. product data are not in dictionary format', lcm.output)

    def test_005__uid2names(self):
        """ test uid2names malformed input """
        data_ele = {'foo': 'bar'}
        self.assertFalse({}, self.productgroup._uid2names(data_ele))

    def test_006__group(self):
        """ test _group() malformed input """
        data_ele = {}
        self.assertFalse(self.productgroup._group(data_ele))

    def test_007__group(self):
        """ test _group() """
        data_ele = {'attributes': {'productGroups': {'foo': {'index': 0, 'name': 'foo', 'products': {'product1': {'uid1': {'index': 1}, 'uid2': {'index': 0}}}}}}}
        self.assertEqual([{'name': 'foo', 'product_list': {1: 'uid1', 0: 'uid2'}}], self.productgroup._group(data_ele))

    def test_008__group(self):
        """ test _group() """
        data_ele = {'attributes':
                    {'productGroups':
                                   {'foo': {'index': 0,
                                            'name': 'foo',
                                            'products': {
                                                'product1': {'uid1': {'index': 1}, 'uid2': {'index': 2}},
                                                'product2': {'uid3': {'index': 0}, 'uid4': {'index': 3}}
                                                }}}}}
        self.assertEqual([{'name': 'foo', 'product_list': {0: 'uid3', 1: 'uid1', 2: 'uid2', 3: 'uid4'}}], self.productgroup._group(data_ele))

    def test_009__group(self):
        """ test _group() """
        data_ele = {'attributes':
                        {'productGroups':
                            {'foo': {'index': 1, 'name': 'foo',
                                    'products': {
                                        'product1': {'uid1': {'index': 1}, 'uid2': {'index': 2}},
                                        'product2': {'uid3': {'index': 0}, 'uid4': {'index': 3}}
                                    }},
                            'bar': {'index': 0, 'name': 'bar',
                                    'products': {
                                        'product1': {'uid4': {'index': 1}, 'uid5': {'index': 2}},
                                        'product2': {'uid6': {'index': 0}, 'uid7': {'index': 3}}
                                    }}

                                    }}}
        result = [{'name': 'bar', 'product_list': {1: 'uid4', 2: 'uid5', 0: 'uid6', 3: 'uid7'}}, {'name': 'foo', 'product_list': {1: 'uid1', 2: 'uid2', 0: 'uid3', 3: 'uid4'}}]
        self.assertEqual(result, self.productgroup._group(data_ele))

    @patch('dkb_robo.portfolio.ProductGroup._group')
    @patch('dkb_robo.portfolio.ProductGroup._uid2names')
    def test_010_map(self, mock_uid2names, mock_group):
        """ test map() """
        data = {'data': {'foo': 'bar'}}
        mock_uid2names.return_value = 'mock_uid2names'
        mock_group.return_value = 'mock_group'
        self.assertEqual(('mock_uid2names', 'mock_group'), self.productgroup.map(data))
        self.assertTrue(mock_uid2names.called)
        self.assertTrue(mock_group.called)


class TestOverview(unittest.TestCase):
    """ test class """

    @patch('requests.Session')
    def setUp(self, mock_session):
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        self.overview = Overview(client=mock_session)
        self.maxDiff = None

    def test_011__fetch(self):
        """ test _fetch() """
        self.overview.client = Mock()
        self.overview.client.get.return_value.status_code = 200
        self.overview.client.get.return_value.json.return_value = {'foo': 'bar'}
        self.assertEqual({'foo': 'bar'}, self.overview._fetch('url'))
        self.assertTrue(self.overview.client.get.called)

    def test_012__fetch(self):
        """ test _fetch() """
        self.overview.client = Mock()
        self.overview.client.get.return_value.status_code = 400
        self.overview.client.get.return_value.json.return_value = {'foo': 'bar'}
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertFalse(self.overview._fetch('url'))
        self.assertIn('ERROR:dkb_robo.portfolio:fetch url: RC is not 200 but 400', lcm.output)
        self.assertTrue(self.overview.client.get.called)

    @patch('dkb_robo.portfolio.Overview._sort')
    @patch('dkb_robo.portfolio.Overview._fetch')
    def test_013_get(self, mock_fetch, mock_sort):
        """ test get() """
        mock_fetch.return_value = {'data': 'data'}
        mock_sort.return_value = 'mock_sort'
        self.assertEqual('mock_sort', self.overview.get())
        self.assertTrue(mock_fetch.called)
        self.assertTrue(mock_sort.called)

    @patch('dkb_robo.portfolio.Overview._sort')
    @patch('dkb_robo.portfolio.Overview._fetch')
    def test_014_get(self, mock_fetch, mock_sort):
        """ test get() """
        mock_fetch.return_value = None
        mock_sort.return_value = 'mock_sort'
        self.assertEqual('mock_sort', self.overview.get())
        self.assertTrue(mock_fetch.called)
        self.assertTrue(mock_sort.called)

    @patch('dkb_robo.portfolio.AccountItem', autospec=True)
    @patch('dkb_robo.portfolio.CardItem', autospec=True)
    @patch('dkb_robo.portfolio.DepotItem', autospec=True)
    def test_015_itemize(self, mock_depot, mock_card, mock_account):
        """ test _itemize() formatted """
        portfolio_dic = {
            'accounts': {'data': [{'id': 'acc1', 'attributes': {'amount': 100}}]},
            'cards': {'data': [{'id': 'card1', 'attributes': {'limit': 500}}]},
            'depots': {'data': [{'id': 'depot1', 'attributes': {'value': 1000}}]}
        }

        mock_account.return_value.format.return_value = 'formatted_account'
        mock_card.return_value.format.return_value = 'formatted_card'
        mock_depot.return_value.format.return_value = 'formatted_depot'

        result = self.overview._itemize(portfolio_dic)

        self.assertEqual(result, {
            'acc1': 'formatted_account',
            'card1': 'formatted_card',
            'depot1': 'formatted_depot'
        })

    @patch('dkb_robo.portfolio.AccountItem', autospec=True)
    @patch('dkb_robo.portfolio.CardItem', autospec=True)
    @patch('dkb_robo.portfolio.DepotItem', autospec=True)
    def test_016_itemize(self, mock_depot, mock_card, mock_account):
        """ test _itemize() unfiltered """
        portfolio_dic = {
            'accounts': {'data': [{'id': 'acc1', 'attributes': {'amount': 100}}]},
            'cards': {'data': [{'id': 'card1', 'attributes': {'limit': 500}}]},
            'depots': {'data': [{'id': 'depot1', 'attributes': {'value': 1000}}]}
        }

        mock_account.return_value = 'unfiltered_account'
        mock_card.return_value = 'unfiltered_card'
        mock_depot.return_value = 'unfiltered_depot'
        self.overview.unfiltered = True
        result = self.overview._itemize(portfolio_dic)

        self.assertEqual(result, {
            'acc1': 'unfiltered_account',
            'card1': 'unfiltered_card',
            'depot1': 'unfiltered_depot'
        })

    @patch('dkb_robo.portfolio.ProductGroup', autospec=True)
    @patch('dkb_robo.portfolio.Overview._itemize', autospec=True)
    @patch('dkb_robo.portfolio.Overview._add_remaining', autospec=True)
    def test_017_sort(self, mock_add, mock_itemize, mock_pgrp):
        """ test _sort() formatted """
        portfolio_dic = {
            'product_display': {'data': [{'id': 'display1'}]},
            'accounts': {'data': [{'id': 'acc1', 'attributes': {'amount': 100}}]},
            'cards': {'data': [{'id': 'card1', 'attributes': {'limit': 500}}]},
            'depots': {'data': [{'id': 'depot1', 'attributes': {'value': 1000}}]}
        }

        mock_itemize.return_value = {
            'acc1': {'amount': 100},
            'card1': {'limit': 500},
            'depot1': {'value': 1000}
        }

        mock_product_group = MagicMock()
        mock_product_group.map.return_value = ({'acc1': 'Account'}, [{'name': 'Group1', 'product_list': {'acc1': 'acc1'}}])
        mock_pgrp.return_value = mock_product_group

        mock_add.return_value = {
            0: {'amount': 100, 'productgroup': 'Group1'}
        }

        result = self.overview._sort(portfolio_dic)

        self.assertEqual(result, {
            0: {'amount': 100, 'productgroup': 'Group1'}
        })

    @patch('dkb_robo.portfolio.ProductGroup', autospec=True)
    @patch('dkb_robo.portfolio.Overview._itemize', autospec=True)
    @patch('dkb_robo.portfolio.Overview._add_remaining', autospec=True)
    def test_018_sort(self, mock_add, mock_itemize, mock_pgrp):
        """ test _sort() unfiltered """
        portfolio_dic = {
            'product_display': {'data': [{'id': 'display1'}]},
            'accounts': {'data': [{'id': 'acc1', 'attributes': {'amount': 100}}]},
            'cards': {'data': [{'id': 'card1', 'attributes': {'limit': 500}}]},
            'depots': {'data': [{'id': 'depot1', 'attributes': {'value': 1000}}]}
        }

        mock_itemize.return_value = {
            'acc1': MagicMock(),
            'card1': MagicMock(),
            'depot1': MagicMock()
        }

        mock_product_group = MagicMock()
        mock_product_group.map.return_value = ({'acc1': 'Account'}, [{'name': 'Group1', 'product_list': {'acc1': 'acc1'}}])
        mock_pgrp.return_value = mock_product_group

        mock_add.return_value = {
            0: {'amount': 100, 'productgroup': 'Group1'}
        }

        self.overview.unfiltered = True
        result = self.overview._sort(portfolio_dic)

        self.assertEqual(result, {
            0: {'amount': 100, 'productgroup': 'Group1'}
        })

    def test_019_add_remaining(self):
        """ test _add_remaining() formatted """
        data_dic = {
            'product1': {'name': 'Product 1'},
            'product2': {'name': 'Product 2'}
        }
        account_dic = {}
        account_cnt = 0

        result = self.overview._add_remaining(data_dic, account_dic, account_cnt)

        expected_result = {
            0: {'name': 'Product 1', 'productgroup': None},
            1: {'name': 'Product 2', 'productgroup': None}
        }

        self.assertEqual(result, expected_result)

    def test_020_add_remaining(self):
        """ test _add_remaining() unfiltered """
        self.overview.unfiltered = True
        data_dic = {
            'product1': MagicMock(),
            'product2': MagicMock()
        }
        account_dic = {}
        account_cnt = 0

        result = self.overview._add_remaining(data_dic, account_dic, account_cnt)

        data_dic['product1'].productGroup = None
        data_dic['product2'].productGroup = None

        expected_result = {
            0: data_dic['product1'],
            1: data_dic['product2']
        }

        self.assertEqual(result, expected_result)

    @patch('dkb_robo.portfolio.Overview._fetch')
    def test_021_get(self, mock_fetch):
        """ test get() e2e """
        pd_dic = json_load(self.dir_path + '/mocks/pd.json')
        account_dic = json_load(self.dir_path + '/mocks/accounts.json')
        card_dic = json_load(self.dir_path + '/mocks/cards.json')
        depot_dic = json_load(self.dir_path + '/mocks/brokerage.json')
        mock_fetch.side_effect = [pd_dic, account_dic, card_dic, depot_dic, {'foo': 'bar'}]

        result = {
                0: {'account': '987654321',
                    'amount': 1234.56,
                    'currencyCode': 'EUR',
                    'holderName': 'HolderName1',
                    'id': 'baccountid1',
                    'name': 'pdsettings brokeraage baccountid1',
                    'productgroup': 'productGroup name 1',
                    'transactions': 'https://banking.dkb.de/api/broker/brokerage-accounts/baccountid1/positions?include=instrument%2Cquote',
                    'type': 'depot'},
                1: {'account': 'AccountIBAN3',
                    'amount': -1000.22,
                    'currencyCode': 'EUR',
                    'date': '2020-03-01',
                    'holderName': 'Account HolderName 3',
                    'iban': 'AccountIBAN3',
                    'id': 'accountid3',
                    'limit': 2500.0,
                    'name': 'pdsettings accoutname accountid3',
                    'productgroup': 'productGroup name 1',
                    'transactions': 'https://banking.dkb.de/api/accounts/accounts/accountid3/transactions',
                    'type': 'account'},
                2: {'account': 'maskedPan1',
                    'amount': -1234.56,
                    'currencycode': 'EUR',
                    'date': '2020-01-03',
                    'expirydate': '2020-01-02',
                    'holdername:': 'holderfirstName holderlastName',
                    'id': 'cardid1',
                    'limit': 1000.0,
                    'maskedpan': 'maskedPan1',
                    'name': 'pdsettings cardname cardid1',
                    'productgroup': 'productGroup name 1',
                    'status': {'category': 'active',
                                'final': None,
                                'limitationsFor': [],
                                'reason': None,
                                'since': None},
                    'transactions': 'https://banking.dkb.de/api/card-transactions/creditcard-transactions?cardId=cardid1',
                    'type': 'creditcard'},
                3: {'account': 'maskedPan2',
                    'amount': 12345.67,
                    'currencycode': 'EUR',
                    'date': '2020-02-07',
                    'expirydate': '2020-02-02',
                    'holdername:': 'holderfirstName2 holderlastName2',
                    'id': 'cardid2',
                    'limit': 0.0,
                    'maskedpan': 'maskedPan2',
                    'name': 'displayName2',
                    'productgroup': 'productGroup name 1',
                    'status': {'category': 'active',
                                'final': None,
                                'limitationsFor': [],
                                'reason': None,
                                'since': None},
                    'transactions': 'https://banking.dkb.de/api/card-transactions/creditcard-transactions?cardId=cardid2',
                    'type': 'creditcard'},
                4: {'account': 'AccountIBAN2',
                    'amount': 1284.56,
                    'currencyCode': 'EUR',
                    'date': '2020-02-01',
                    'holderName': 'Account HolderName 2',
                    'iban': 'AccountIBAN2',
                    'id': 'accountid2',
                    'limit': 0.0,
                    'name': 'pdsettings accoutname accountid2',
                    'productgroup': 'productGroup name 2',
                    'transactions': 'https://banking.dkb.de/api/accounts/accounts/accountid2/transactions',
                    'type': 'account'},
                5: {'account': 'AccountIBAN1',
                    'amount': 12345.67,
                    'currencyCode': 'EUR',
                    'date': '2020-01-01',
                    'holderName': 'Account HolderName 1',
                    'iban': 'AccountIBAN1',
                    'id': 'accountid1',
                    'limit': 1000.0,
                    'name': 'Account DisplayName 1',
                    'productgroup': None,
                    'transactions': 'https://banking.dkb.de/api/accounts/accounts/accountid1/transactions',
                    'type': 'account'},
                6: {'account': 'maskedPan3',
                    'expirydate': '2020-04-04',
                    'holdername:': 'holderfirstName3 holderlastName3',
                    'id': 'cardid3',
                    'limit': None,
                    'maskedpan': 'maskedPan3',
                    'name': 'Visa Debitkarte',
                    'productgroup': None,
                    'status': {'category': 'blocked',
                                'final': True,
                                'limitationsFor': None,
                                'reason': 'cancellation-of-product-by-customer',
                                'since': '2020-03-01'},
                    'transactions': None,
                    'type': 'debitcard'},
                7: {'account': 'maskedPan4',
                    'expirydate': '2020-04-03',
                    'holdername:': 'holderfirstName4 holderlastName4',
                    'id': 'cardid4',
                    'limit': None,
                    'maskedpan': 'maskedPan4',
                    'name': 'Visa Debitkarte',
                    'productgroup': None,
                    'status': {'category': 'active',
                                'final': None,
                                'limitationsFor': None,
                                'reason': None,
                                'since': None},
                    'transactions': None,
                    'type': 'debitcard'}}

        self.assertEqual(result, self.overview.get())

class TestAccountItem(unittest.TestCase):
    """ test class """
    def setUp(self):
        """ setup """
        self.amount_data = {'value': 1000, 'currencyCode': 'USD'}
        self.interests_data = [
            {
                'details': [
                    {'condition': {'currency': 'USD', 'maximumAmount': 5000, 'minimumAmount': 1000}, 'interestRate': 1.5}
                ],
                'method': 'simple',
                'type': 'savings'
            }
        ]
        self.product_data = {'id': 'prod123', 'type': 'savings', 'displayName': 'Savings Account'}
        self.account_data = {
            'availableBalance': self.amount_data,
            'balance': self.amount_data,
            'currencyCode': 'USD',
            'displayName': 'My Account',
            'holderName': 'John Doe',
            'id': 'acc123',
            'iban': 'DE1234567890',
            'interestRate': {'value': 1.5, 'currencyCode': 'USD'},
            'interests': self.interests_data,
            'lastAccountStatementDate': '2022-01-01',
            'nearTimeBalance': self.amount_data,
            'openingDate': '2020-01-01',
            'overdraftLimit': 500.0,
            'permissions': ['view', 'edit'],
            'product': self.product_data,
            'productGroup': 'group1',
            'state': 'active',
            'type': 'checking',
            'unauthorizedOverdraftInterestRate': {'value': 2.5, 'currencyCode': 'USD'},
            'updatedAt': '2022-01-15'
        }

    def test_022_post_init(self):
        """ test post init """
        self.maxDiff = None
        account = AccountItem(**self.account_data)
        self.assertEqual(account.id, 'acc123')
        self.assertEqual(account.displayName, 'My Account')
        self.assertEqual(account.holderName, 'John Doe')
        self.assertEqual(account.iban, 'DE1234567890')
        self.assertEqual(account.currencyCode, 'USD')
        self.assertEqual(account.type, 'checking')
        self.assertEqual(account.state, 'active')
        self.assertEqual(account.permissions, ['view', 'edit'])
        self.assertEqual(account.productGroup, 'group1')
        self.assertEqual(account.openingDate, '2020-01-01')
        self.assertEqual(account.lastAccountStatementDate, '2022-01-01')
        self.assertEqual(account.updatedAt, '2022-01-15')
        self.assertEqual(account.balance.value, 1000)
        self.assertEqual(account.balance.currencyCode, 'USD')
        self.assertEqual(account.availableBalance.value, 1000)
        self.assertEqual(account.availableBalance.currencyCode, 'USD')
        self.assertEqual(account.nearTimeBalance.value, 1000)
        self.assertEqual(account.nearTimeBalance.currencyCode, 'USD')
        self.assertEqual(account.overdraftLimit, 500.0)
        self.assertEqual(account.interestRate['value'], 1.5)
        self.assertEqual(account.interestRate['currencyCode'], 'USD')
        self.assertEqual(account.interests[0].type, 'savings')
        self.assertEqual(account.interests[0].method, 'simple')
        self.assertEqual(account.interests[0].details[0].condition.currency, 'USD')
        self.assertEqual(account.interests[0].details[0].interestRate, 1.5)
        self.assertEqual(account.interests[0].details[0].condition.maximumAmount, 5000)
        self.assertEqual(account.interests[0].details[0].condition.minimumAmount, 1000)
        self.assertEqual(account.unauthorizedOverdraftInterestRate['value'], 2.5)
        self.assertEqual(account.product.id, 'prod123')
        self.assertEqual(account.product.type, 'savings')
        self.assertEqual(account.product.displayName, 'Savings Account')

    def test_023_post_init(self):
        """ test post init with invalid data """
        self.maxDiff = None
        self.account_data['overdraftLimit'] = 'aa'
        self.account_data['interests'][0]['details'][0]['condition']['maximumAmount'] = 'aa'
        self.account_data['interests'][0]['details'][0]['condition']['minimumAmount'] = 'aa'
        account = AccountItem(**self.account_data)
        self.assertEqual(account.id, 'acc123')
        self.assertEqual(account.displayName, 'My Account')
        self.assertEqual(account.holderName, 'John Doe')
        self.assertEqual(account.iban, 'DE1234567890')
        self.assertEqual(account.currencyCode, 'USD')
        self.assertEqual(account.type, 'checking')
        self.assertEqual(account.state, 'active')
        self.assertEqual(account.permissions, ['view', 'edit'])
        self.assertEqual(account.productGroup, 'group1')
        self.assertEqual(account.openingDate, '2020-01-01')
        self.assertEqual(account.lastAccountStatementDate, '2022-01-01')
        self.assertEqual(account.updatedAt, '2022-01-15')
        self.assertEqual(account.balance.value, 1000)
        self.assertEqual(account.balance.currencyCode, 'USD')
        self.assertEqual(account.availableBalance.value, 1000)
        self.assertEqual(account.availableBalance.currencyCode, 'USD')
        self.assertEqual(account.nearTimeBalance.value, 1000)
        self.assertEqual(account.nearTimeBalance.currencyCode, 'USD')
        self.assertEqual(account.overdraftLimit, None)
        self.assertEqual(account.interestRate['value'], 1.5)
        self.assertEqual(account.interestRate['currencyCode'], 'USD')
        self.assertEqual(account.interests[0].type, 'savings')
        self.assertEqual(account.interests[0].method, 'simple')
        self.assertEqual(account.interests[0].details[0].condition.currency, 'USD')
        self.assertEqual(account.interests[0].details[0].interestRate, 1.5)
        self.assertEqual(account.interests[0].details[0].condition.maximumAmount, None)
        self.assertEqual(account.interests[0].details[0].condition.minimumAmount, None)
        self.assertEqual(account.unauthorizedOverdraftInterestRate['value'], 2.5)
        self.assertEqual(account.product.id, 'prod123')
        self.assertEqual(account.product.type, 'savings')
        self.assertEqual(account.product.displayName, 'Savings Account')

    def test_024_post_init(self):
        """ test post init """
        self.maxDiff = None
        self.account_data['nearTimeBalance'] = None
        account = AccountItem(**self.account_data)
        self.assertEqual(account.id, 'acc123')
        self.assertEqual(account.displayName, 'My Account')
        self.assertEqual(account.holderName, 'John Doe')
        self.assertEqual(account.iban, 'DE1234567890')
        self.assertEqual(account.currencyCode, 'USD')
        self.assertEqual(account.type, 'checking')
        self.assertEqual(account.state, 'active')
        self.assertEqual(account.permissions, ['view', 'edit'])
        self.assertEqual(account.productGroup, 'group1')
        self.assertEqual(account.openingDate, '2020-01-01')
        self.assertEqual(account.lastAccountStatementDate, '2022-01-01')
        self.assertEqual(account.updatedAt, '2022-01-15')
        self.assertEqual(account.balance.value, 1000)
        self.assertEqual(account.balance.currencyCode, 'USD')
        self.assertEqual(account.availableBalance.value, 1000)
        self.assertEqual(account.availableBalance.currencyCode, 'USD')
        self.assertFalse(account.nearTimeBalance)
        self.assertEqual(account.overdraftLimit, 500.0)
        self.assertEqual(account.interestRate['value'], 1.5)
        self.assertEqual(account.interestRate['currencyCode'], 'USD')
        self.assertEqual(account.interests[0].type, 'savings')
        self.assertEqual(account.interests[0].method, 'simple')
        self.assertEqual(account.interests[0].details[0].condition.currency, 'USD')
        self.assertEqual(account.interests[0].details[0].interestRate, 1.5)
        self.assertEqual(account.interests[0].details[0].condition.maximumAmount, 5000)
        self.assertEqual(account.interests[0].details[0].condition.minimumAmount, 1000)
        self.assertEqual(account.unauthorizedOverdraftInterestRate['value'], 2.5)
        self.assertEqual(account.product.id, 'prod123')
        self.assertEqual(account.product.type, 'savings')
        self.assertEqual(account.product.displayName, 'Savings Account')

    @patch('dkb_robo.portfolio.logger')
    def test_024_format(self, mock_logger):
        """ test format() """
        account = AccountItem(**self.account_data)
        formatted_account = account.format()

        expected_output = {
            'account': 'DE1234567890',
            'amount': 1000,
            'currencyCode': 'USD',
            'date': '2022-01-15',
            'holderName': 'John Doe',
            'iban': 'DE1234567890',
            'id': 'acc123',
            'limit': 500.0,
            'name': 'Savings Account',
            'transactions': "https://banking.dkb.de/api/accounts/accounts/acc123/transactions",
            'type': 'checking',
        }
        self.assertEqual(formatted_account, expected_output)


class TestCardItem(unittest.TestCase):
    """ test class """
    def setUp(self):
        """ setup """
        self.maxDiff = None
        self.amount_data = {'value': 1000, 'currencyCode': 'USD'}
        self.person_data = {'firstName': 'John', 'lastName': 'Doe'}
        self.limit_data = {'value': 5000, 'currencyCode': 'USD', 'identifier': 'limit1', 'categories': [{'amount': self.amount_data, 'name': 'category1'}]}
        self.product_data = {'superProductId': 'prod123', 'displayName': 'Credit Card', 'institute': 'Bank', 'productType': 'credit', 'ownerType': 'individual', 'id': 'prod123', 'type': 'creditCard'}
        self.status_data = {'category': 'active', 'since': '2022-01-01', 'reason': 'none', 'final': True, 'limitationsFor': []}
        self.card_data = {
            'activationDate': '2022-01-01',
            'authorizedAmount': {'value': 1000, 'currencyCode': 'USD'},
            'availableLimit': {'value': 1000, 'currencyCode': 'USD'},
            'balance': self.amount_data,
            'billingDetails': {'days': [1, 15], 'calendarType': 'monthly', 'cycle': 'monthly'},
            'blockedSince': '2022-01-01',
            'creationDate': '2022-01-01',
            'engravedLine1': 'John Doe',
            'engravedLine2': 'Credit Card',
            'expiryDate': '2025-01-01',
            'failedPinAttempts': 0,
            'followUpCardId': 'card123',
            'holder': {'person': self.person_data},
            'id': 'card123',
            'limit': self.limit_data,
            'maskedPan': '1234********5678',
            'network': 'Visa',
            'owner': self.person_data,
            'product': self.product_data,
            'referenceAccount': {'iban': 'DE1234567890', 'bic': 'BICCODE'},
            'state': 'active',
            'status': self.status_data,
            'type': 'creditCard'
        }

    def test_025_post_init(self):
        """ test post init """
        card = CardItem(**self.card_data)
        self.assertEqual(card.id, 'card123')
        self.assertEqual(card.activationDate, '2022-01-01')
        self.assertEqual(card.authorizedAmount.value, 1000)
        self.assertEqual(card.authorizedAmount.currencyCode, 'USD')
        self.assertEqual(card.availableLimit.value, 1000)
        self.assertEqual(card.availableLimit.currencyCode, 'USD')
        self.assertEqual(card.balance.value, 1000)
        self.assertEqual(card.balance.currencyCode, 'USD')
        self.assertEqual(card.billingDetails.days, [1, 15])
        self.assertEqual(card.billingDetails.calendarType, 'monthly')
        self.assertEqual(card.billingDetails.cycle, 'monthly')
        self.assertEqual(card.blockedSince, '2022-01-01')
        self.assertEqual(card.creationDate, '2022-01-01')
        self.assertEqual(card.engravedLine1, 'John Doe')
        self.assertEqual(card.engravedLine2, 'Credit Card')
        self.assertEqual(card.expiryDate, '2025-01-01')
        self.assertEqual(card.failedPinAttempts, 0)
        self.assertEqual(card.followUpCardId, 'card123')
        self.assertEqual(card.holder.person.firstName, 'John')
        self.assertEqual(card.holder.person.lastName, 'Doe')
        self.assertEqual(card.limit.value, 5000)
        self.assertEqual(card.limit.currencyCode, 'USD')
        self.assertEqual(card.limit.identifier, 'limit1')
        self.assertEqual(card.limit.categories[0].amount.value, 1000)
        self.assertEqual(card.limit.categories[0].amount.currencyCode, 'USD')
        self.assertEqual(card.limit.categories[0].name, 'category1')
        self.assertEqual(card.maskedPan, '1234********5678')
        self.assertEqual(card.network, 'Visa')
        self.assertEqual(card.owner.firstName, 'John')
        self.assertEqual(card.owner.lastName, 'Doe')
        self.assertEqual(card.product.superProductId, 'prod123')
        self.assertEqual(card.product.displayName, 'Credit Card')
        self.assertEqual(card.product.institute, 'Bank')
        self.assertEqual(card.product.productType, 'credit')
        self.assertEqual(card.product.ownerType, 'individual')
        self.assertEqual(card.product.id, 'prod123')
        self.assertEqual(card.product.type, 'creditCard')
        self.assertEqual(card.referenceAccount.iban, 'DE1234567890')
        self.assertEqual(card.referenceAccount.bic, 'BICCODE')
        self.assertEqual(card.state, 'active')
        self.assertEqual(card.status.category, 'active')
        self.assertEqual(card.status.since, '2022-01-01')
        self.assertEqual(card.status.reason, 'none')
        self.assertEqual(card.status.final, True)
        self.assertEqual(card.status.limitationsFor, [])
        self.assertEqual(card.type, 'creditCard')

    def test_026_post_init(self):
        """ test post init with invalid data """
        self.card_data['limit']['value'] = 'aa'
        self.card_data['limit']['categories'][0]['amount']['value'] = 'aa'
        card = CardItem(**self.card_data)
        self.assertEqual(card.limit.value, None)
        self.assertEqual(card.availableLimit.value, 1000)
        self.assertEqual(card.authorizedAmount.value, 1000)
        self.assertEqual(card.limit.categories[0].amount.value, None)
        self.assertEqual(card.id, 'card123')
        self.assertEqual(card.activationDate, '2022-01-01')
        self.assertEqual(card.authorizedAmount.currencyCode, 'USD')
        self.assertEqual(card.availableLimit.currencyCode, 'USD')
        # self.assertEqual(card.balance.value, 1000)
        self.assertEqual(card.balance.currencyCode, 'USD')
        self.assertEqual(card.billingDetails.days, [1, 15])
        self.assertEqual(card.billingDetails.calendarType, 'monthly')
        self.assertEqual(card.billingDetails.cycle, 'monthly')
        self.assertEqual(card.blockedSince, '2022-01-01')
        self.assertEqual(card.creationDate, '2022-01-01')
        self.assertEqual(card.engravedLine1, 'John Doe')
        self.assertEqual(card.engravedLine2, 'Credit Card')
        self.assertEqual(card.expiryDate, '2025-01-01')
        self.assertEqual(card.failedPinAttempts, 0)
        self.assertEqual(card.followUpCardId, 'card123')
        self.assertEqual(card.holder.person.firstName, 'John')
        self.assertEqual(card.holder.person.lastName, 'Doe')

        self.assertEqual(card.limit.currencyCode, 'USD')
        self.assertEqual(card.limit.identifier, 'limit1')
        self.assertEqual(card.limit.categories[0].amount.currencyCode, 'USD')
        self.assertEqual(card.limit.categories[0].name, 'category1')
        self.assertEqual(card.maskedPan, '1234********5678')
        self.assertEqual(card.network, 'Visa')
        self.assertEqual(card.owner.firstName, 'John')
        self.assertEqual(card.owner.lastName, 'Doe')
        self.assertEqual(card.product.superProductId, 'prod123')
        self.assertEqual(card.product.displayName, 'Credit Card')
        self.assertEqual(card.product.institute, 'Bank')
        self.assertEqual(card.product.productType, 'credit')
        self.assertEqual(card.product.ownerType, 'individual')
        self.assertEqual(card.product.id, 'prod123')
        self.assertEqual(card.product.type, 'creditCard')
        self.assertEqual(card.referenceAccount.iban, 'DE1234567890')
        self.assertEqual(card.referenceAccount.bic, 'BICCODE')
        self.assertEqual(card.state, 'active')
        self.assertEqual(card.status.category, 'active')
        self.assertEqual(card.status.since, '2022-01-01')
        self.assertEqual(card.status.reason, 'none')
        self.assertEqual(card.status.final, True)
        self.assertEqual(card.status.limitationsFor, [])
        self.assertEqual(card.type, 'creditCard')

    @patch('dkb_robo.portfolio.logger')
    def test_027_format(self, mock_logger):
        """ test format() for creditcard """
        card = CardItem(**self.card_data)
        formatted_card = card.format()

        expected_output = {
            'account': '1234********5678',
            'expirydate': '2025-01-01',
            'holdername:': 'John Doe',
            'id': 'card123',
            'name': 'Credit Card',
            'limit': 5000,
            'maskedpan': '1234********5678',
            'status': {
                'category': 'active',
                'since': '2022-01-01',
                'reason': 'none',
                'final': True,
                'limitationsFor': []
            },
            'type': 'creditcard',
            'transactions': "https://banking.dkb.de/api/card-transactions/creditcard-transactions?cardId=card123",
            'amount': -1000,
            'currencycode': 'USD',
            'date': None  # Assuming balance.date is None
        }

        self.assertEqual(formatted_card, expected_output)

    @patch('dkb_robo.portfolio.logger')
    def test_028_format(self, mock_logger):
        """ test format() for debitcard """
        self.card_data['type'] = 'debitCard'
        card = CardItem(**self.card_data)
        formatted_card = card.format()

        expected_output = {
            'account': '1234********5678',
            'expirydate': '2025-01-01',
            'holdername:': 'John Doe',
            'id': 'card123',
            'name': 'Credit Card',
            'limit': 5000.0,
            'maskedpan': '1234********5678',
            'status': {
                'category': 'active',
                'since': '2022-01-01',
                'reason': 'none',
                'final': True,
                'limitationsFor': []
            },
            'type': 'debitcard',
            'transactions': None,
        }

        self.assertEqual(formatted_card, expected_output)


class TestDepotItem(unittest.TestCase):
    """ test class """
    def setUp(self):
        """ setup """
        self.amount_data = {'value': 1000, 'currencyCode': 'USD'}
        self.person_data = {'firstName': 'John', 'lastName': 'Doe'}
        self.brokerage_account_performance_data = {
            'currentValue': self.amount_data,
            'averagePrice': self.amount_data,
            'overallAbsolute': self.amount_data,
            'overallRelative': '10%',
            'isOutdated': False
        }
        self.reference_account_data = [
            {'internalReferenceAccounts': False, 'accountType': 'savings', 'accountNumber': '123456', 'bankCode': 'BANK123', 'holderName': 'John Doe'}
        ]
        self.depot_data = {
            'brokerageAccountPerformance': self.brokerage_account_performance_data,
            'depositAccountId': 'acc123',
            'holder': self.person_data,
            'holderName': 'John Doe',
            'id': 'depot123',
            'referenceAccounts': self.reference_account_data,
            'riskClasses': ['low', 'medium'],
            'tradingEnabled': True,
            'type': 'depot'
        }

    def test_029_post_init(self):
        """ test post init """
        depot = DepotItem(**self.depot_data)
        self.assertEqual(depot.id, 'depot123')
        self.assertEqual(depot.depositAccountId, 'acc123')
        self.assertEqual(depot.holder.firstName, 'John')
        self.assertEqual(depot.holder.lastName, 'Doe')
        self.assertEqual(depot.holderName, 'John Doe')
        self.assertEqual(depot.type, 'depot')
        self.assertEqual(depot.tradingEnabled, True)
        self.assertEqual(depot.brokerageAccountPerformance.currentValue.value, 1000)
        self.assertEqual(depot.brokerageAccountPerformance.currentValue.currencyCode, 'USD')
        self.assertEqual(depot.brokerageAccountPerformance.averagePrice.value, 1000)
        self.assertEqual(depot.brokerageAccountPerformance.averagePrice.currencyCode, 'USD')
        self.assertEqual(depot.brokerageAccountPerformance.overallAbsolute.value, 1000)
        self.assertEqual(depot.brokerageAccountPerformance.overallAbsolute.currencyCode, 'USD')
        self.assertEqual(depot.brokerageAccountPerformance.overallRelative, '10%')
        self.assertEqual(depot.brokerageAccountPerformance.isOutdated, False)
        self.assertEqual(depot.referenceAccounts[0].internalReferenceAccounts, False)
        self.assertEqual(depot.referenceAccounts[0].accountType, 'savings')
        self.assertEqual(depot.referenceAccounts[0].accountNumber, '123456')
        self.assertEqual(depot.referenceAccounts[0].bankCode, 'BANK123')
        self.assertEqual(depot.referenceAccounts[0].holderName, 'John Doe')
        self.assertEqual(depot.riskClasses, ['low', 'medium'])

    @patch('dkb_robo.portfolio.logger')
    def test_030_format(self, mock_logger):
        """ test format() """
        depot = DepotItem(**self.depot_data)
        formatted_depot = depot.format()

        expected_output = {
            'account': 'acc123',
            'amount': 1000,
            'currencyCode': 'USD',
            'holderName': 'John Doe',
            'id': 'depot123',
            'name': 'John Doe',
            'type': 'depot',
            'transactions': 'https://banking.dkb.de/api/broker/brokerage-accounts/depot123/positions?include=instrument%2Cquote',
        }

        self.assertEqual(formatted_depot, expected_output)



if __name__ == '__main__':

    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" unittests for dkb_robo """
import sys
import os
import unittest
from pathlib import Path
from unittest.mock import patch, MagicMock, Mock, mock_open
from bs4 import BeautifulSoup
from mechanicalsoup import LinkNotFoundError
from datetime import date
import io
import json

sys.path.insert(0, '.')
sys.path.insert(0, '..')
from dkb_robo import DKBRobo
import logging


def json_load(fname):
    """ simple json load """

    with open(fname, 'r', encoding='utf8') as myfile:
        data_dic = json.load(myfile)

    return data_dic

def read_file(fname):
    """ read file into string """
    with open(fname, "rb") as myfile:
        data = myfile.read()

    return data

def cnt_list(value):
    """ customized function return just the number if entries in input list """
    return len(value)

def my_side_effect(*args, **kwargs):
    """ my sideeffect funtion """
    return [200, args[1], [args[4]]]


class TestDKBRobo(unittest.TestCase):
    """ test class """

    maxDiff = None

    def setUp(self):
        self.dkb = DKBRobo()
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        self.logger = logging.getLogger('dkb_robo')

    @patch('dkb_robo.authentication.Authentication.login')
    @patch('dkb_robo.legacy.Wrapper.login')
    def test_001__enter(self, mock_legacy, mock_api):
        """ test enter """
        mock_legacy.return_value = ('legacy', 'foo')
        mock_api.return_value = ('api', 'foo')
        self.assertTrue(self.dkb.__enter__())
        self.assertFalse(mock_legacy.called)
        self.assertTrue(mock_api.called)

    @patch('dkb_robo.authentication.Authentication.login')
    @patch('dkb_robo.legacy.Wrapper.login')
    def test_002__enter(self, mock_legacy, mock_api):
        """ test enter """
        self.dkb.legacy_login = False
        mock_legacy.return_value = ('legacy', 'foo')
        mock_api.return_value = ('api', 'foo')
        self.assertTrue(self.dkb.__enter__())
        self.assertFalse(mock_legacy.called)
        self.assertTrue(mock_api.called)

    @patch('dkb_robo.authentication.Authentication.login')
    @patch('dkb_robo.legacy.Wrapper.login')
    def test_003__enter(self, mock_legacy, mock_api):
        """ test enter """
        self.dkb.legacy_login = True
        mock_legacy.return_value = ('legacy', 'foo')
        mock_api.return_value = ('api', 'foo')
        with self.assertRaises(Exception) as err:
            self.dkb.__enter__()
        self.assertEqual('Legacy Login got deprecated. Please do not use this option anymore', str(err.exception))
        self.assertFalse(mock_legacy.called)
        self.assertFalse(mock_api.called)

    @patch('dkb_robo.authentication.Authentication.login')
    @patch('dkb_robo.legacy.Wrapper.login')
    def test_004__enter(self, mock_legacy, mock_api):
        """ test enter """
        self.dkb.tan_insert = True
        # self.dkb.wrapper = Mock()
        mock_legacy.return_value = ('legacy', 'foo')
        mock_api.return_value = ('api', 'foo')
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.dkb.__enter__()
        self.assertIn("INFO:dkb_robo:tan_insert is a legacy login option and will be disabled soon. Please use chip_tan instead", lcm.output)
        self.assertTrue(self.dkb.chip_tan)
        self.assertFalse(mock_legacy.called)
        self.assertTrue(mock_api.called)

    @patch('dkb_robo.authentication.Authentication.login')
    @patch('dkb_robo.legacy.Wrapper.login')
    def test_005__enter(self, mock_legacy, mock_api):
        """ test enter """
        self.dkb.mfa_device = 1
        # self.dkb.wrapper = Mock()
        mock_legacy.return_value = ('legacy', 'foo')
        mock_api.return_value = ('api', 'foo')
        self.assertTrue(self.dkb.__enter__())
        self.assertFalse(mock_legacy.called)
        self.assertTrue(mock_api.called)
        self.assertEqual(1, self.dkb.mfa_device)

    @patch('dkb_robo.authentication.Authentication.login')
    @patch('dkb_robo.legacy.Wrapper.login')
    def test_006__enter(self, mock_legacy, mock_api):
        """ test enter """
        self.dkb.mfa_device = 2
        # self.dkb.wrapper = Mock()
        mock_legacy.return_value = ('legacy', 'foo')
        mock_api.return_value = ('api', 'foo')
        self.assertTrue(self.dkb.__enter__())
        self.assertFalse(mock_legacy.called)
        self.assertTrue(mock_api.called)
        self.assertEqual(2, self.dkb.mfa_device)

    @patch('dkb_robo.authentication.Authentication.login')
    @patch('dkb_robo.legacy.Wrapper.login')
    def test_007__enter(self, mock_legacy, mock_api):
        """ test enter """
        self.dkb.mfa_device = 'm'
        # self.dkb.wrapper = Mock()
        mock_legacy.return_value = ('legacy', 'foo')
        mock_api.return_value = ('api', 'foo')
        self.assertTrue(self.dkb.__enter__())
        self.assertFalse(mock_legacy.called)
        self.assertTrue(mock_api.called)
        self.assertEqual(1, self.dkb.mfa_device)

    def test_008__exit(self):
        """ test enter """
        self.dkb.wrapper = Mock()
        self.dkb.wrapper.logout = Mock()
        self.assertFalse(self.dkb.__exit__())
        self.assertTrue(self.dkb.wrapper.logout.called)

    @patch('dkb_robo.transaction.Transactions.get')
    @patch('dkb_robo.dkb_robo.validate_dates')
    def test_009_get_transactions(self, mock_date, mock_transaction):
        """ test get_transactions() """
        self.dkb.legacy_login = True
        mock_date.return_value = ('from', 'to')
        self.dkb.wrapper = Mock()
        mock_transaction.return_value = {'foo': 'bar'}
        self.assertEqual({'foo': 'bar'}, self.dkb.get_transactions('url', 'account', 'from', 'to', 'btype'))

    @patch('dkb_robo.transaction.Transactions.get')
    @patch('dkb_robo.dkb_robo.validate_dates')
    def test_010_get_transactions(self, mock_date, mock_transaction):
        """ test get_transactions() """
        self.dkb.legacy_login = False
        mock_date.return_value = ('from', 'to')
        self.dkb.wrapper = Mock()
        mock_transaction.return_value = {'foo': 'bar'}
        self.assertEqual({'foo': 'bar'}, self.dkb.get_transactions('url', 'atype', 'from', 'to', 'btype'))

    def test_011_get_points(self):
        """ test get_exemption_order()"""
        self.dkb.wrapper = Mock()
        with self.assertRaises(Exception) as err:
            self.dkb.get_points()
        self.assertEqual('Method not supported...', str(err.exception))

    @patch('dkb_robo.exemptionorder.ExemptionOrders.fetch')
    def test_012_get_exemption_order(self, mock_fetch):
        """ test get_exemption_order()"""
        self.dkb.wrapper = Mock()
        mock_fetch.return_value = {'foo': 'bar'}
        self.assertEqual({'foo': 'bar'}, self.dkb.get_exemption_order())

    def test_013_get_credit_limits(self):
        """ test get_credit_limits()"""
        self.dkb.account_dic = {1: {'limit': 1000, 'foo': 'bar', 'iban': 'iban'}, 2: {'limit': 2000, 'foo': 'bar', 'maskedpan': 'maskedpan'}}
        self.assertEqual({'iban': 1000, 'maskedpan': 2000}, self.dkb.get_credit_limits())

    @patch('dkb_robo.standingorder.StandingOrders.fetch')
    def test_014_get_standing_orders(self, mock_fetch):
        """ test get_standing_orders()"""
        self.dkb.wrapper = Mock()
        mock_fetch.return_value = {'foo': 'bar'}
        self.assertEqual({'foo': 'bar'}, self.dkb.get_standing_orders())

    def test_015_scan_postbox(self):
        """ test get_standing_orders()"""
        self.dkb.download = Mock()
        self.dkb.download.return_value = {'foo': 'bar'}
        self.assertEqual({'foo': 'bar'}, self.dkb.scan_postbox())

    @patch('dkb_robo.dkb_robo.DKBRobo.format_doc', autospec=True)
    @patch('dkb_robo.postbox.PostBox.fetch_items')
    @patch('dkb_robo.dkb_robo.DKBRobo.download_doc', autospec=True)
    def test_016_download(self, mock_download_doc, mock_fetch_items, mock_format):
        """ download_all_documents"""
        path = Path('/some/path')
        self.dkb.wrapper = Mock()
        self.dkb.wrapper.client = Mock()
        self.dkb.wrapper.account_dic = {'id1': {'id': 'id1', 'account': 'account1', 'read': False, 'foo': 'bar', 'iban': 'iban'}, 'id2': {'id': 'id2', 'account': 'account2', 'read': True, 'foo': 'bar'}}
        unread_doc = MagicMock()
        unread_doc.message.read = False
        read_doc = MagicMock()
        read_doc.message.read = True
        mock_fetch_items.return_value = {
            'doc1': unread_doc,
            'doc2': read_doc
        }

        mock_format.return_value = {
            'doc1': 'document1',
            'doc2': 'document2'
        }

        documents = self.dkb.download(path=path, download_all=False)
        self.assertEqual({'doc1': 'document1', 'doc2': 'document2'}, documents)
        self.assertTrue(mock_format.called)
        self.assertTrue(mock_fetch_items.called)

    @patch('dkb_robo.dkb_robo.DKBRobo.format_doc', autospec=True)
    @patch('dkb_robo.postbox.PostBox.fetch_items')
    @patch('dkb_robo.dkb_robo.DKBRobo.download_doc', autospec=True)
    def test_017_download(self, mock_download_doc, mock_fetch_items, mock_format):
        """ download_all_documents"""
        path = Path('/some/path')
        self.dkb.wrapper = Mock()
        self.dkb.wrapper.client = Mock()
        self.dkb.wrapper.account_dic = {'id1': {'id': 'id1', 'account': 'account1', 'read': False, 'foo': 'bar', 'iban': 'iban'}, 'id2': {'id': 'id2', 'account': 'account2', 'read': True, 'foo': 'bar'}}
        mock_fetch_items.return_value = {
            'doc1': 'document1',
            'doc2': 'document2'
        }

        mock_format.return_value = {
            'doc1': 'document1',
            'doc2': 'document2'
        }

        documents = self.dkb.download(path=path, download_all=True)
        self.assertEqual({'doc1': 'document1', 'doc2': 'document2'}, documents)
        self.assertTrue(mock_format.called)
        self.assertTrue(mock_fetch_items.called)
        mock_download_doc.assert_any_call(self.dkb, path=path, doc=documents['doc1'], prepend_date=False, mark_read=True, use_account_folders=False, list_only=False, accounts_by_id={'id1': 'account1', 'id2': 'account2'})
        mock_download_doc.assert_any_call(self.dkb, path=path, doc=documents['doc2'], prepend_date=False, mark_read=True, use_account_folders=False, list_only=False, accounts_by_id={'id1': 'account1', 'id2': 'account2'})


    @patch('dkb_robo.dkb_robo.DKBRobo.format_doc')
    @patch('dkb_robo.postbox.PostBox.fetch_items')
    @patch('dkb_robo.dkb_robo.DKBRobo.download_doc', autospec=True)
    def test_018_download(self, mock_download_doc, mock_fetch_items, mock_format):
        """ download_all_documents"""
        path = Path('/some/path')
        self.dkb.wrapper = Mock()
        self.dkb.wrapper.client = Mock()
        self.dkb.wrapper.account_dic = {'id1': {'id': 'id1', 'account': 'account1', 'read': False, 'foo': 'bar', 'iban': 'iban'}, 'id2': {'id': 'id2', 'account': 'account2', 'read': True, 'foo': 'bar'}}
        mock_format.return_value = {
            'doc1': 'document1',
            'doc2': 'document2'
        }

        documents = self.dkb.download(path=None, download_all=True)
        self.assertEqual({'doc1': 'document1', 'doc2': 'document2'}, documents)
        mock_download_doc.assert_not_called()
        self.assertTrue(mock_format.called)
        self.assertTrue(mock_fetch_items.called)

    @patch('dkb_robo.dkb_robo.time.sleep', autospec=True)
    def test_019_download_doc(self, mock_sleep):
        """ download a single document """
        path = Path('/some/path')
        doc = MagicMock()
        doc.category.return_value = 'category'
        doc.account.return_value = 'account'
        doc.date.return_value = '2022-01-01'
        doc.filename.return_value = 'document.pdf'
        doc.subject.return_value = 'Document Subject'
        doc.download.return_value = True
        self.dkb.wrapper = MagicMock()
        self.dkb.download_doc(path=path, doc=doc, prepend_date=True, mark_read=True, use_account_folders=True, list_only=False, accounts_by_id={})
        target = path / 'category' / 'account'
        filename = '2022-01-01_document.pdf'
        doc.download.assert_called_with(self.dkb.wrapper.client, target / filename)
        doc.mark_read.assert_called_with(self.dkb.wrapper.client, True)
        mock_sleep.assert_called_once_with(0.5)

    def test_020_download_doc(self):
        """ list only   """
        path = Path('/some/path')
        doc = MagicMock()
        doc.category.return_value = 'category'
        doc.account.return_value = 'account'
        doc.date.return_value = '2022-01-01'
        doc.filename.return_value = 'document.pdf'
        doc.subject.return_value = 'Document Subject'
        doc.download.return_value = True
        self.dkb.logger = MagicMock()
        self.dkb.wrapper = MagicMock()
        self.dkb.download_doc(path=path, doc=doc, prepend_date=True, mark_read=True, use_account_folders=True, list_only=True, accounts_by_id={})
        self.dkb.logger.info.assert_not_called()
        doc.download.assert_not_called()
        doc.mark_read.assert_not_called()

    @patch('dkb_robo.dkb_robo.time.sleep', autospec=True)
    def test_021_download_doc(self, mock_sleep):
        """ test existing file """
        path = Path('/some/path')
        doc = MagicMock()
        doc.category.return_value = 'category'
        doc.account.return_value = 'account'
        doc.date.return_value = '2022-01-01'
        doc.filename.return_value = 'document.pdf'
        doc.subject.return_value = 'Document Subject'
        doc.download.return_value = False
        self.dkb.logger = MagicMock()
        self.dkb.wrapper = MagicMock()
        self.dkb.download_doc(path=path, doc=doc, prepend_date=True, mark_read=True, use_account_folders=True, list_only=False, accounts_by_id={})
        target = path / 'category' / 'account'
        filename = '2022-01-01_document.pdf'
        self.dkb.logger.info.assert_called_with("File already exists. Skipping %s.", filename)
        doc.download.assert_called_with(self.dkb.wrapper.client, target / filename)
        doc.mark_read.assert_not_called()
        mock_sleep.assert_not_called()

    def test_022_format_doc(self):
        """ format documents """
        self.path = Path('/some/path')
        doc1 = MagicMock()
        doc1.category.return_value = 'category1'
        doc1.account.return_value = 'account1'
        doc1.date.return_value = '2022-01-01'
        doc1.filename.return_value = 'document1.pdf'
        doc1.message.subject = 'Subject1'
        doc1.id = 'id1'
        doc1.document.creationDate = '2022-01-01'
        doc1.document.link = 'link1'
        doc1.rcode = 'rcode1'

        doc2 = MagicMock()
        doc2.category.return_value = 'category2'
        doc2.account.return_value = 'account2'
        doc2.date.return_value = '2022-01-02'
        doc2.filename.return_value = 'document2.pdf'
        doc2.message.subject = 'Subject2'
        doc2.id = 'id2'
        doc2.document.creationDate = '2022-01-02'
        doc2.document.link = 'link2'
        doc2.rcode = 'rcode2'

        documents = {'doc1': doc1, 'doc2': doc2}
        accounts_by_id = {'account1': 'Account 1', 'account2': 'Account 2'}

        result = self.dkb.format_doc(self.path, documents, use_account_folders=True, prepend_date=True, accounts_by_id=accounts_by_id)

        expected_result = {
            'category1': {
                'documents': {
                    'Subject1': {
                        'id': 'id1',
                        'date': '2022-01-01',
                        'link': 'link1',
                        'fname': str(self.path / 'category1' / 'account1' / '2022-01-01_document1.pdf'),
                        'rcode': 'rcode1'
                    }
                },
                'count': 1
            },
            'category2': {
                'documents': {
                    'Subject2': {
                        'id': 'id2',
                        'date': '2022-01-02',
                        'link': 'link2',
                        'fname': str(self.path / 'category2' / 'account2' / '2022-01-02_document2.pdf'),
                        'rcode': 'rcode2'
                    }
                },
                'count': 1
            }
        }

        self.assertEqual(result, expected_result)

    def test_023_format_doc(self):
        """ format documents without account folders """
        self.path = Path('/some/path')
        doc1 = MagicMock()
        doc1.category.return_value = 'category1'
        doc1.date.return_value = '2022-01-01'
        doc1.filename.return_value = 'document1.pdf'
        doc1.message.subject = 'Subject1'
        doc1.id = 'id1'
        doc1.document.creationDate = '2022-01-01'
        doc1.document.link = 'link1'
        doc1.rcode = 'rcode1'
        documents = {'doc1': doc1}
        result = self.dkb.format_doc(self.path, documents, use_account_folders=False, prepend_date=False)

        expected_result = {
            'category1': {
                'documents': {
                    'Subject1': {
                        'id': 'id1',
                        'date': '2022-01-01',
                        'link': 'link1',
                        'fname': str(self.path / 'category1' / 'document1.pdf'),
                        'rcode': 'rcode1'
                    }
                },
                'count': 1
            }
        }

        self.assertEqual(result, expected_result)

    def test_024_accounts_by_id(self):
        """ test accounts_by_id unfiltered True """
        self.dkb.unfiltered = True
        self.dkb.wrapper = MagicMock()
        self.dkb.wrapper.account_dic = {
            '1': MagicMock(id='1', iban='DE1234567890', maskedPan=None, depositAccountId=None),
            '2': MagicMock(id='2', iban=None, maskedPan='1234********5678', depositAccountId=None),
            '3': MagicMock(id='3', iban=None, maskedPan=None, depositAccountId='DEPO123456')
        }

        result = self.dkb._accounts_by_id()
        expected_result = {
            '1': 'DE1234567890',
            '2': '1234********5678',
            '3': 'DEPO123456'
        }
        self.assertEqual(result, expected_result)

    def test_ß25_accounts_by_id(self):
        """ test accounts_by_id unfiltered False """
        self.dkb.unfiltered = False
        self.dkb.wrapper = MagicMock()
        self.dkb.wrapper.account_dic = {
            '1': {'id': '1', 'account': 'DE1234567890'},
            '2': {'id': '2', 'account': '1234********5678'},
            '3': {'id': '3', 'account': 'DEPO123456'}
        }
        result = self.dkb._accounts_by_id()
        expected_result = {
            '1': 'DE1234567890',
            '2': '1234********5678',
            '3': 'DEPO123456'
        }

        self.assertEqual(result, expected_result)





if __name__ == '__main__':

    unittest.main()



# -*- coding: utf-8 -*-
# pylint: disable=r0904, c0415, c0413, r0913, w0212
""" unittests for dkb_robo """
import sys
import os
from datetime import date
import unittest
import logging
import json
from unittest.mock import patch, Mock, MagicMock, mock_open
from bs4 import BeautifulSoup
from mechanicalsoup import LinkNotFoundError
import io
sys.path.insert(0, '.')
sys.path.insert(0, '..')
from dkb_robo.authentication import Authentication, APPAuthentication, TANAuthentication

def json_load(fname):
    """ simple json load """

    with open(fname, 'r', encoding='utf8') as myfile:
        data_dic = json.load(myfile)

    return data_dic


class TestAuthentication(unittest.TestCase):
    """ test class """

    def setUp(self):
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        self.logger = logging.getLogger('dkb_robo')
        self.auth = Authentication()
        self.maxDiff = None

    def test_001_init(self):
        """ test init() """
        self.auth.__init__()
        self.assertEqual('seal_one', self.auth.mfa_method)

    def test_002_init(self):
        """ test init() """
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.auth.__init__(chip_tan=True)
        self.assertIn('INFO:dkb_robo.authentication:Using to chip_tan to login', lcm.output)
        self.assertEqual('chip_tan_manual', self.auth.mfa_method)

    def test_003_init(self):
        """ test init() """
        self.auth.__init__(chip_tan=False)
        self.assertEqual('seal_one', self.auth.mfa_method)

    def test_004_init(self):
        """ test init() """
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.auth.__init__(chip_tan='qr')
        self.assertIn('INFO:dkb_robo.authentication:Using to chip_tan to login', lcm.output)
        self.assertEqual('chip_tan_qr', self.auth.mfa_method)

    def test_005_init(self):
        """ test init() """
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.auth.__init__(chip_tan='chip_tan_qr')
        self.assertIn('INFO:dkb_robo.authentication:Using to chip_tan to login', lcm.output)
        self.assertEqual('chip_tan_qr', self.auth.mfa_method)

    @patch('requests.session')
    def test_006__session_new(self, mock_session):
        """ test _session_new() """
        mock_session.headers = {}
        client = self.auth._session_new()
        exp_headers = {'Accept-Language': 'en-US,en;q=0.5', 'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8', 'Cache-Control': 'no-cache', 'Connection': 'keep-alive', 'DNT': '1', 'Pragma': 'no-cache', 'Sec-Fetch-Dest': 'document', 'Sec-Fetch-Mode': 'navigate', 'Sec-Fetch-Site': 'none', 'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:128.0) Gecko/20100101 Firefox/128.0', 'priority': 'u=0', 'sec-gpc': '1', 'te': 'trailers'}
        self.assertEqual(exp_headers, client.headers)

    @patch('requests.session')
    def test_007__session_new(self, mock_session):
        """ test _session_new() """
        mock_session.headers = {}
        self.auth.proxies = 'proxies'
        client = self.auth._session_new()
        self.assertEqual('proxies', client.proxies)

    @patch('requests.session')
    def test_008__session_new(self, mock_session):
        """ test _session_new() """
        mock_session.headers = {}
        mock_session.get.return_value.status_code = 200
        mock_session.return_value.cookies = {'__Host-xsrf': 'foo'}
        client = self.auth._session_new()
        exp_headers = {'Accept-Language': 'en-US,en;q=0.5', 'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8', 'Cache-Control': 'no-cache', 'Connection': 'keep-alive', 'DNT': '1', 'Pragma': 'no-cache', 'Sec-Fetch-Dest': 'document', 'Sec-Fetch-Mode': 'navigate', 'Sec-Fetch-Site': 'none', 'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:128.0) Gecko/20100101 Firefox/128.0', 'x-xsrf-token': 'foo', 'priority': 'u=0', 'sec-gpc': '1', 'te': 'trailers'}
        self.assertEqual(exp_headers, client.headers)

    def test_009_token_get(self):
        """ test _token_get() ok """
        self.auth.dkb_user = 'dkb_user'
        self.auth.dkb_password = 'dkb_password'
        self.auth.client = Mock()
        self.auth.client.post.return_value.status_code = 200
        self.auth.client.post.return_value.json.return_value = {'foo': 'bar'}
        self.auth._token_get()
        self.assertEqual({'foo': 'bar'}, self.auth.token_dic)

    def test_010_token_get(self):
        """ test _token_get() ok """
        self.auth.dkb_user = 'dkb_user'
        self.auth.dkb_password = 'dkb_password'
        self.auth.client = Mock()
        self.auth.client.post.return_value.status_code = 400
        self.auth.client.post.return_value.json.return_value = {'foo': 'bar'}
        with self.assertRaises(Exception) as err:
            self.auth._token_get()
        self.assertEqual('Login failed: 1st factor authentication failed. RC: 400', str(err.exception))
        self.assertFalse(self.auth.token_dic)

    def test_011__mfa_get(self):
        """ test _mfa_get() """
        self.auth.token_dic = {'foo': 'bar'}
        with self.assertRaises(Exception) as err:
            self.auth._mfa_get()
        self.assertEqual('Login failed: no 1fa access token.', str(err.exception))

    def test_012__mfa_get(self):
        """ test _mfa_get() """
        self.auth.token_dic = {'access_token': 'bar', 'mfa_id': 'mfa_id'}
        self.auth.client = Mock()
        self.auth.client.get.return_value.status_code = 400
        with self.assertRaises(Exception) as err:
            self.auth._mfa_get()
        self.assertEqual('Login failed: getting mfa_methods failed. RC: 400', str(err.exception))

    def test_013__mfa_get(self):
        """ test _mfa_get() """
        self.auth.client = Mock()
        self.auth.client.get.return_value.status_code = 200
        self.auth.client.get.return_value.json.return_value = {'foo1': 'bar1'}
        self.auth.token_dic = {'access_token': 'bar', 'mfa_id': 'mfa_id'}
        self.assertEqual({'foo1': 'bar1'}, self.auth._mfa_get())

    def test_014__mfa_sort(self):
        """ test sort_mfa_devices() """
        mfa_dic = {
            'data': [
                {'attributes': {'preferredDevice': False, 'enrolledAt': '2022-01-01'}},
                {'attributes': {'preferredDevice': True, 'enrolledAt': '2022-01-02'}},
                {'attributes': {'preferredDevice': False, 'enrolledAt': '2022-01-03'}}
            ]
        }
        expected_result = {
            'data': [
                {'attributes': {'preferredDevice': True, 'enrolledAt': '2022-01-02'}},
                {'attributes': {'preferredDevice': False, 'enrolledAt': '2022-01-01'}},
                {'attributes': {'preferredDevice': False, 'enrolledAt': '2022-01-03'}}
            ]
        }
        self.assertEqual(expected_result, self.auth._mfa_sort(mfa_dic))


    def test_015__mfa_sort(self):
        """ test sort_mfa_devices() """
        mfa_dic = {
            'data': [
                {'attributes': {'preferredDevice': False, 'enrolledAt': '2022-01-03'}},
                {'attributes': {'preferredDevice': True, 'enrolledAt': '2022-01-02'}},
                {'attributes': {'preferredDevice': False, 'enrolledAt': '2022-01-01'}}
            ]
        }
        expected_result = {
            'data': [
                {'attributes': {'preferredDevice': True, 'enrolledAt': '2022-01-02'}},
                {'attributes': {'preferredDevice': False, 'enrolledAt': '2022-01-01'}},
                {'attributes': {'preferredDevice': False, 'enrolledAt': '2022-01-03'}}
            ]
        }
        self.assertEqual(expected_result, self.auth._mfa_sort(mfa_dic))


    def test_016__mfa_sort(self):
        """ test sort_mfa_devices() """
        mfa_dic = {
            'data': [
                {'attributes': {'preferredDevice': False, 'enrolledAt': '2022-01-03'}},
                {'attributes': {'preferredDevice': False, 'enrolledAt': '2022-01-02'}},
                {'attributes': {'preferredDevice': True, 'enrolledAt': '2022-01-04'}}
            ]
        }
        expected_result = {
            'data': [
                {'attributes': {'preferredDevice': True, 'enrolledAt': '2022-01-04'}},
                {'attributes': {'preferredDevice': False, 'enrolledAt': '2022-01-02'}},
                {'attributes': {'preferredDevice': False, 'enrolledAt': '2022-01-03'}}
            ]
        }
        self.assertEqual(expected_result, self.auth._mfa_sort(mfa_dic))

    def test_017__mfa_select(self):
        """ test _mfa_select() """
        mfa_dic = {'foo': 'bar'}
        self.auth.mfa_device = 1
        self.assertEqual(0, self.auth._mfa_select(mfa_dic))

    def test_018__mfa_select(self):
        """ test _mfa_select() """
        mfa_dic = {'foo': 'bar'}
        self.auth.mfa_device = 2
        self.assertEqual(1, self.auth._mfa_select(mfa_dic))

    def test_019__mfa_select(self):
        """ test _mfa_select() """
        mfa_dic = {'foo': 'bar'}
        self.assertEqual(0, self.auth._mfa_select(mfa_dic))

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    @patch('builtins.input')
    def test_020__mfa_select(self, mock_input, mock_stdout):
        """ test _mfa_select() """
        mock_input.return_value=1
        self.auth.mfa_device = 0
        mfa_dic = {'data': [{'attributes': {'deviceName': 'device-1'}}, {'attributes': {'deviceName': 'device-2'}}]}
        self.assertEqual(0, self.auth._mfa_select(mfa_dic))
        self.assertIn("\nPick an authentication device from the below list:\n[1] - device-1\n[2] - device-2\n", mock_stdout.getvalue())

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    @patch('builtins.input')
    def test_021__mfa_select(self, mock_input, mock_stdout):
        """ test _mfa_select() """
        mock_input.return_value=1
        self.auth.mfa_device = 4
        mfa_dic = {'data': [{'attributes': {'deviceName': 'device-1'}}, {'attributes': {'deviceName': 'device-2'}}]}
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(0, self.auth._mfa_select(mfa_dic))
        self.assertIn('WARNING:dkb_robo.authentication:User submitted mfa_device number is invalid. Ingoring...', lcm.output)
        self.assertIn("\nPick an authentication device from the below list:\n[1] - device-1\n[2] - device-2\n", mock_stdout.getvalue())

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    @patch('builtins.input')
    def test_022__mfa_select(self, mock_input, mock_stdout):
        """ test _mfa_select() """
        mock_input.return_value=1
        mfa_dic = {'data': [{'attributes': {'deviceName': 'device-1'}}, {'attributes': {'deviceName': 'device-2'}}]}
        self.assertEqual(0, self.auth._mfa_select(mfa_dic))
        self.assertIn("\nPick an authentication device from the below list:\n[1] - device-1\n[2] - device-2\n", mock_stdout.getvalue())

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    @patch('builtins.input')
    def test_023__mfa_select(self, mock_input, mock_stdout):
        """ test _mfa_select() """
        mock_input.return_value=2
        mfa_dic = {'data': [{'attributes': {'deviceName': 'device-1'}}, {'attributes': {'deviceName': 'device-2'}}]}
        self.assertEqual(1, self.auth._mfa_select(mfa_dic))
        self.assertIn("\nPick an authentication device from the below list:\n[1] - device-1\n[2] - device-2\n", mock_stdout.getvalue())

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    @patch('builtins.input')
    def test_024__mfa_select(self, mock_input, mock_stdout):
        """ test _mfa_select() """
        mock_input.side_effect = [4, 1]
        mfa_dic = {'data': [{'attributes': {'deviceName': 'device-1'}}, {'attributes': {'deviceName': 'device-2'}}]}
        self.assertEqual(0, self.auth._mfa_select(mfa_dic))
        self.assertIn("\nPick an authentication device from the below list:\n[1] - device-1\n[2] - device-2\n", mock_stdout.getvalue())
        self.assertIn('Wrong input!', mock_stdout.getvalue())

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    @patch('builtins.input')
    def test_025__mfa_select(self, mock_input, mock_stdout):
        """ test _mfa_select() """
        mock_input.side_effect = ['a', 4, 1]
        mfa_dic = {'data': [{'attributes': {'deviceName': 'device-1'}}, {'attributes': {'deviceName': 'device-2'}}]}
        self.assertEqual(0, self.auth._mfa_select(mfa_dic))
        self.assertIn("\nPick an authentication device from the below list:\n[1] - device-1\n[2] - device-2\n", mock_stdout.getvalue())
        self.assertIn('Invalid input!', mock_stdout.getvalue())
        self.assertIn('Wrong input!', mock_stdout.getvalue())

    @patch('requests.session')
    def test_026__mfa_challenge(self, mock_session):
        """ test _mfa_challenge() """
        mfa_dic = {}
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(({}, None), self.auth._mfa_challenge(mfa_dic, 1))
        self.assertIn('ERROR:dkb_robo.authentication:mfa dictionary has an unexpected data structure', lcm.output)

    @patch('requests.session')
    def test_027__mfa_challenge(self, mock_session):
        """ test _mfa_challenge() """
        self.auth.client = Mock()
        self.auth.client.headers = {}
        self.auth.client.post.return_value.status_code = 200
        self.auth.client.post.return_value.json.return_value = {'data': {'type': 'mfa-challenge', 'id': 'id'}}
        self.auth.client.headers = {'foo': 'bar'}
        self.auth.mfa_method = 'seal_one'
        self.auth.token_dic = {'mfa_id': 'mfa_id'}
        mfa_dic = {'data': {1: {'id': 'id', 'attributes': {'deviceName': 'deviceName'}}}}
        self.assertEqual(({'data': {'id': 'id', 'type': 'mfa-challenge'}}, 'deviceName'), self.auth._mfa_challenge(mfa_dic, 1))

    @patch('requests.session')
    def test_028__mfa_challenge(self, mock_session):
        """ test _mfa_challenge() """
        self.auth.client = Mock()
        self.auth.client.headers = {}
        self.auth.client.post.return_value.status_code = 400
        self.auth.client.post.return_value.json.return_value = {'data': {'type': 'mfa-challenge', 'id': 'id'}}
        self.auth.client.headers = {'foo': 'bar'}
        self.auth.mfa_method = 'seal_one'
        self.auth.token_dic = {'mfa_id': 'mfa_id'}
        mfa_dic = {'data': {1: {'id': 'id', 'attributes': {'deviceName': 'deviceName'}}}}
        with self.assertRaises(Exception) as err:
            self.assertEqual(({'data': {'id': 'id', 'type': 'mfa-challenge'}}, 'deviceName'), self.auth._mfa_challenge(mfa_dic, 1))
        self.assertEqual('Login failed: post request to get the mfa challenges failed. RC: 400', str(err.exception))

    @patch('requests.session')
    def test_029__mfa_challenge(self, mock_session):
        """ test _mfa_challenge() """
        self.auth.client = Mock()
        self.auth.client.headers = {}
        self.auth.client.post.return_value.status_code = 200
        self.auth.client.post.return_value.json.return_value = {'data': {'type': 'mfa-challenge', 'id': 'id'}}
        self.auth.client.headers = {'foo': 'bar'}
        self.auth.mfa_method = 'seal_one'
        self.auth.token_dic = {'mfa_id': 'mfa_id'}
        mfa_dic = {'data': {1: {'id': 'id', 'attributes': {'foo': 'bar'}}}}
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(({'data': {'id': 'id', 'type': 'mfa-challenge'}}, None), self.auth._mfa_challenge(mfa_dic, 1))
        self.assertIn('ERROR:dkb_robo.authentication:unable to get mfa-deviceName for device_num: 1', lcm.output)

    @patch('dkb_robo.authentication.TANAuthentication.finalize')
    @patch('dkb_robo.authentication.APPAuthentication.finalize')
    @patch('dkb_robo.authentication.Authentication._mfa_challenge_id')
    def test_030__mfa_finalize(self, mock_cid, mock_app, mock_ctm):
        """ test _mfa_finalize() """
        mock_cid.return_value = 'cid'
        mock_app.return_value = 'app'
        mock_ctm.return_value = 'ctm'
        self.auth.mfa_method = 'seal_one'
        self.assertEqual('app', self.auth._mfa_finalize('challenge_dic', 'device_name'))
        self.assertTrue(mock_cid.called)
        self.assertTrue(mock_app.called)
        self.assertFalse(mock_ctm.called)

    @patch('dkb_robo.authentication.TANAuthentication.finalize')
    @patch('dkb_robo.authentication.APPAuthentication.finalize')
    @patch('dkb_robo.authentication.Authentication._mfa_challenge_id')
    def test_031__mfa_finalize(self, mock_cid, mock_app, mock_ctm):
        """ test _mfa_finalize() """
        mock_cid.return_value = 'cid'
        mock_app.return_value = 'app'
        mock_ctm.return_value = 'ctm'
        self.auth.mfa_method = 'chip_tan_manual'
        self.assertEqual('ctm', self.auth._mfa_finalize('challenge_dic', 'device_name'))
        self.assertTrue(mock_cid.called)
        self.assertFalse(mock_app.called)
        self.assertTrue(mock_ctm.called)

    @patch('dkb_robo.authentication.TANAuthentication.finalize')
    @patch('dkb_robo.authentication.APPAuthentication.finalize')
    @patch('dkb_robo.authentication.Authentication._mfa_challenge_id')
    def test_032__mfa_finalize(self, mock_cid, mock_app, mock_ctm):
        """ test _mfa_finalize() """
        mock_cid.return_value = 'cid'
        mock_app.return_value = 'app'
        mock_ctm.return_value = 'ctm'
        self.auth.mfa_method = 'other_method'
        with self.assertRaises(Exception) as err:
            self.assertFalse(self.auth._mfa_finalize('challenge_dic', 'device_name'))
        self.assertEqual('Login failed: unknown mfa method: other_method', str(err.exception))
        self.assertTrue(mock_cid.called)
        self.assertFalse(mock_app.called)
        self.assertFalse(mock_ctm.called)

    @patch('requests.session')
    def test_033__mfa_challenge_id(self, mock_session):
        """ test _mfa_challenge_id() """
        mfa_dic = {}
        with self.assertRaises(Exception) as err:
            self.assertFalse(self.auth._mfa_challenge_id(mfa_dic))
        self.assertEqual('Login failed: challenge response format is other than expected: {}', str(err.exception))

    @patch('requests.session')
    def test_034__mfa_challenge_id(self, mock_session):
        """ test _mfa_challenge_id() """
        mfa_dic = {'data': {'id': 'id', 'type': 'type'}}
        with self.assertRaises(Exception) as err:
            self.assertFalse(self.auth._mfa_challenge_id(mfa_dic))
        self.assertEqual("Login failed:: wrong challenge type: {'data': {'id': 'id', 'type': 'type'}}", str(err.exception))

    @patch('requests.session')
    def test_035__mfa_challenge_id(self, mock_session):
        """ test _mfa_challenge_id() """
        mfa_dic = {'data': {'id': 'id', 'type': 'mfa-challenge'}}
        self.assertEqual('id', self.auth._mfa_challenge_id(mfa_dic))

    def test_036__token_update(self):
        """ test _token_update() ok """
        self.auth.token_dic = {'mfa_id': 'mfa_id', 'access_token': 'access_token'}
        self.auth.client = Mock()
        self.auth.client.post.return_value.status_code = 200
        self.auth.client.post.return_value.json.return_value = {'foo': 'bar'}
        self.auth._token_update()
        self.assertEqual({'foo': 'bar'}, self.auth.token_dic)

    def test_037__token_update(self):
        """ test _token_update() nok """
        self.auth.token_dic = {'mfa_id': 'mfa_id', 'access_token': 'access_token'}
        self.auth.client = Mock()
        self.auth.client.post.return_value.status_code = 400
        self.auth.client.post.return_value.json.return_value = {'foo': 'bar'}
        with self.assertRaises(Exception) as err:
            self.auth._token_update()
        self.assertEqual('Login failed: token update failed. RC: 400', str(err.exception))
        self.assertEqual({'mfa_id': 'mfa_id', 'access_token': 'access_token'}, self.auth.token_dic)

    @patch('dkb_robo.legacy.Wrapper._new_instance')
    def test_037__sso_redirect(self, mock_instance):
        """ test _sso_redirect() ok """
        self.auth.client = Mock()
        self.auth.client.headers = {}
        self.auth.client.post.return_value.status_code = 200
        self.auth.client.post.return_value.text = 'OK'
        self.auth._sso_redirect()
        self.assertTrue(mock_instance.called)

    @patch('dkb_robo.legacy.Wrapper._new_instance')
    def test_038__sso_redirect(self, mock_instance):
        """ test _sso_redirect() nok """
        self.auth.client = Mock()
        self.auth.client.headers = {}
        self.auth.client.post.return_value.status_code = 200
        self.auth.client.post.return_value.text = 'NOK'
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.auth._sso_redirect()
        self.assertIn('ERROR:dkb_robo.authentication:SSO redirect failed. RC: 200 text: NOK', lcm.output)
        self.assertTrue(mock_instance.called)

    @patch('dkb_robo.legacy.Wrapper._new_instance')
    def test_039__sso_redirect(self, mock_instance):
        """ test _sso_redirect() nok """
        self.auth.client = Mock()
        self.auth.client.headers = {}
        self.auth.client.post.return_value.status_code = 400
        self.auth.client.post.return_value.text = 'OK'
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.auth._sso_redirect()
        self.assertIn('ERROR:dkb_robo.authentication:SSO redirect failed. RC: 400 text: OK', lcm.output)
        self.assertTrue(mock_instance.called)

    @patch('dkb_robo.authentication.Authentication._mfa_sort')
    @patch('dkb_robo.authentication.Authentication._mfa_get')
    @patch('dkb_robo.authentication.Authentication._token_get')
    @patch('dkb_robo.authentication.Authentication._session_new')
    def test_040_login(self, mock_sess, mock_tok, mock_meth, mock_sort):
        """ test login() """
        self.auth.token_dic = {'foo': 'bar'}
        mock_meth.return_value = {'foo': 'bar'}
        mock_sort.return_value = {'foo': 'bar'}
        with self.assertRaises(Exception) as err:
            self.auth.login()
        self.assertEqual('Login failed: no 1fa access token.', str(err.exception))
        self.assertTrue(mock_sess.called)
        self.assertTrue(mock_tok.called)
        self.assertTrue(mock_meth.called)

    @patch('dkb_robo.authentication.Authentication._mfa_sort')
    @patch('dkb_robo.authentication.Authentication._mfa_select')
    @patch('dkb_robo.authentication.Authentication._mfa_get')
    @patch('dkb_robo.authentication.Authentication._token_get')
    @patch('dkb_robo.authentication.Authentication._session_new')
    def test_041_login(self, mock_sess, mock_tok, mock_meth, mock_mfa, mock_sort):
        """ test login() """
        self.auth.token_dic = {'mfa_id': 'mfa_id'}
        mock_meth.return_value = {'foo': 'bar'}
        mock_sort.return_value = {'foo': 'bar'}
        mock_mfa.return_value = 0
        with self.assertRaises(Exception) as err:
            self.auth.login()
        self.assertEqual('Login failed: no 1fa access token.', str(err.exception))
        self.assertTrue(mock_sess.called)
        self.assertTrue(mock_tok.called)
        self.assertTrue(mock_meth.called)
        self.assertTrue(mock_mfa.called)
        self.assertTrue(mock_sort.called)

    @patch('dkb_robo.authentication.Authentication._mfa_sort')
    @patch('dkb_robo.authentication.Authentication._mfa_select')
    @patch('dkb_robo.authentication.Authentication._mfa_challenge')
    @patch('dkb_robo.authentication.Authentication._mfa_get')
    @patch('dkb_robo.authentication.Authentication._token_get')
    @patch('dkb_robo.authentication.Authentication._session_new')
    def test_041_login(self, mock_sess, mock_tok, mock_meth, mock_chall, mock_mfa, mock_sort):
        """ test login() """
        self.auth.token_dic = {'mfa_id': 'mfa_id'}
        mock_meth.return_value = {'data': 'bar'}
        mock_sort.return_value = {'data': 'bar'}
        mock_chall.return_value = (None, None)
        mock_mfa.return_value = 0
        with self.assertRaises(Exception) as err:
            self.auth.login()
        self.assertEqual('Login failed: No challenge id.', str(err.exception))
        self.assertTrue(mock_sess.called)
        self.assertTrue(mock_tok.called)
        self.assertTrue(mock_meth.called)
        self.assertTrue(mock_chall.called)
        self.assertTrue(mock_mfa.called)
        self.assertTrue(mock_sort.called)

    @patch('dkb_robo.authentication.Authentication._mfa_sort')
    @patch('dkb_robo.authentication.Authentication._mfa_select')
    @patch('dkb_robo.authentication.Authentication._mfa_finalize')
    @patch('dkb_robo.authentication.Authentication._mfa_challenge')
    @patch('dkb_robo.authentication.Authentication._mfa_get')
    @patch('dkb_robo.authentication.Authentication._token_get')
    @patch('dkb_robo.authentication.Authentication._session_new')
    def test_042_login(self, mock_sess, mock_tok, mock_meth, mock_chall, mock_2fa, mock_mfa, mock_sort):
        """ test login() """
        self.auth.token_dic = {'mfa_id': 'mfa_id'}
        mock_meth.return_value = {'data': 'bar'}
        mock_sort.return_value = {'data': 'bar'}
        mock_chall.return_value = ('mfa_challenge_id', 'deviceName')
        mock_2fa.return_value = False
        mock_mfa.return_value = 0
        with self.assertRaises(Exception) as err:
            self.auth.login()
        self.assertEqual('Login failed: mfa did not complete', str(err.exception))
        self.assertTrue(mock_sess.called)
        self.assertTrue(mock_tok.called)
        self.assertTrue(mock_meth.called)
        self.assertTrue(mock_chall.called)
        self.assertTrue(mock_2fa.called)
        self.assertTrue(mock_mfa.called)
        self.assertTrue(mock_sort.called)

    @patch('dkb_robo.authentication.Authentication._mfa_sort')
    @patch('dkb_robo.authentication.Authentication._mfa_select')
    @patch('dkb_robo.authentication.Authentication._sso_redirect')
    @patch('dkb_robo.authentication.Authentication._token_update')
    @patch('dkb_robo.authentication.Authentication._mfa_finalize')
    @patch('dkb_robo.authentication.Authentication._mfa_challenge')
    @patch('dkb_robo.authentication.Authentication._mfa_get')
    @patch('dkb_robo.authentication.Authentication._token_get')
    @patch('dkb_robo.authentication.Authentication._session_new')
    def test_043_login(self, mock_sess, mock_tok, mock_meth, mock_chall, mock_2fa, mock_upd, mock_redir, mock_mfa, mock_sort):
        """ test login() """
        self.auth.token_dic = {'mfa_id': 'mfa_id'}
        mock_meth.return_value = {'data': 'bar'}
        mock_sort.return_value = {'data': 'bar'}
        mock_chall.return_value = ('mfa_challenge_id', 'deviceName')
        mock_2fa.return_value = True
        mock_mfa.return_value = 0
        with self.assertRaises(Exception) as err:
            self.auth.login()
        self.assertEqual('Login failed: mfa did not complete', str(err.exception))
        self.assertTrue(mock_sess.called)
        self.assertTrue(mock_tok.called)
        self.assertTrue(mock_meth.called)
        self.assertTrue(mock_chall.called)
        self.assertTrue(mock_2fa.called)
        self.assertFalse(mock_redir.called)
        self.assertTrue(mock_mfa.called)
        self.assertTrue(mock_sort.called)

    @patch('dkb_robo.authentication.Authentication._mfa_sort')
    @patch('dkb_robo.authentication.Authentication._mfa_select')
    @patch('dkb_robo.authentication.Authentication._sso_redirect')
    @patch('dkb_robo.authentication.Authentication._token_update')
    @patch('dkb_robo.authentication.Authentication._mfa_finalize')
    @patch('dkb_robo.authentication.Authentication._mfa_challenge')
    @patch('dkb_robo.authentication.Authentication._mfa_get')
    @patch('dkb_robo.authentication.Authentication._token_get')
    @patch('dkb_robo.authentication.Authentication._session_new')
    def test_044_login(self, mock_sess, mock_tok, mock_meth, mock_chall, mock_2fa, mock_upd, mock_redir, mock_mfa, mock_sort):
        """ test login() """
        self.auth.token_dic = {'mfa_id': 'mfa_id', 'access_token': 'access_token'}
        mock_meth.return_value = {'data': 'bar'}
        mock_sort.return_value = {'data': 'bar'}
        mock_chall.return_value = ('mfa_challenge_id', 'deviceName')
        mock_2fa.return_value = True
        mock_mfa.return_value = 0
        with self.assertRaises(Exception) as err:
            self.auth.login()
        self.assertEqual('Login failed: token_factor_type is missing', str(err.exception))
        self.assertTrue(mock_sess.called)
        self.assertTrue(mock_tok.called)
        self.assertTrue(mock_meth.called)
        self.assertTrue(mock_chall.called)
        self.assertTrue(mock_2fa.called)
        self.assertTrue(mock_upd.called)
        self.assertFalse(mock_redir.called)
        self.assertTrue(mock_mfa.called)
        self.assertTrue(mock_sort.called)

    @patch('dkb_robo.authentication.Authentication._mfa_sort')
    @patch('dkb_robo.authentication.Authentication._mfa_select')
    @patch('dkb_robo.authentication.Authentication._sso_redirect')
    @patch('dkb_robo.authentication.Authentication._token_update')
    @patch('dkb_robo.authentication.Authentication._mfa_finalize')
    @patch('dkb_robo.authentication.Authentication._mfa_challenge')
    @patch('dkb_robo.authentication.Authentication._mfa_get')
    @patch('dkb_robo.authentication.Authentication._token_get')
    @patch('dkb_robo.authentication.Authentication._session_new')
    def test_045_login(self, mock_sess, mock_tok, mock_meth, mock_chall, mock_2fa, mock_upd, mock_redir, mock_mfa, mock_sort):
        """ test login() """
        self.auth.token_dic = {'mfa_id': 'mfa_id', 'access_token': 'access_token', 'token_factor_type': 'token_factor_type'}
        mock_meth.return_value = {'data': 'bar'}
        mock_sort.return_value = {'data': 'bar'}
        mock_chall.return_value = ('mfa_challenge_id', 'deviceName')
        mock_mfa.return_value = 0
        mock_2fa.return_value = True
        with self.assertRaises(Exception) as err:
            self.auth.login()
        self.assertEqual('Login failed: 2nd factor authentication did not complete', str(err.exception))
        self.assertTrue(mock_sess.called)
        self.assertTrue(mock_tok.called)
        self.assertTrue(mock_meth.called)
        self.assertTrue(mock_chall.called)
        self.assertTrue(mock_2fa.called)
        self.assertTrue(mock_upd.called)
        self.assertFalse(mock_redir.called)
        self.assertTrue(mock_mfa.called)
        self.assertTrue(mock_sort.called)

    @patch('dkb_robo.authentication.Authentication._mfa_sort')
    @patch('dkb_robo.portfolio.Overview.get')
    @patch('dkb_robo.authentication.Authentication._mfa_select')
    @patch('dkb_robo.authentication.Authentication._sso_redirect')
    @patch('dkb_robo.authentication.Authentication._token_update')
    @patch('dkb_robo.authentication.Authentication._mfa_finalize')
    @patch('dkb_robo.authentication.Authentication._mfa_challenge')
    @patch('dkb_robo.authentication.Authentication._mfa_get')
    @patch('dkb_robo.authentication.Authentication._token_get')
    @patch('dkb_robo.authentication.Authentication._session_new')
    def test_046_login(self, mock_sess, mock_tok, mock_meth, mock_chall, mock_2fa, mock_upd, mock_redir, mock_mfa, mock_overview, mock_sort):
        """ test login() """
        self.auth.token_dic = {'mfa_id': 'mfa_id', 'access_token': 'access_token', 'token_factor_type': '2fa'}
        mock_meth.return_value = {'data': 'bar'}
        mock_sort.return_value = {'data': 'bar'}
        mock_chall.return_value = ('mfa_challenge_id', 'deviceName')
        mock_mfa.return_value = 0
        mock_2fa.return_value = True
        self.auth.login()
        self.assertTrue(mock_sess.called)
        self.assertTrue(mock_tok.called)
        self.assertTrue(mock_meth.called)
        self.assertTrue(mock_chall.called)
        self.assertTrue(mock_2fa.called)
        self.assertTrue(mock_upd.called)
        self.assertTrue(mock_redir.called)
        self.assertTrue(mock_mfa.called)
        self.assertTrue(mock_overview.called)

    def test_147_logout(self):
        """ test logout """
        self.assertFalse(self.auth.logout())


class TestAPPAuthentication(unittest.TestCase):
    """ test class """

    @patch("requests.Session")
    def setUp(self, mock_session):
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        self.appauth = APPAuthentication(client=mock_session)
        # self.maxDiff = None

    def test_001__check(self):
        """ test _check_processing_status() """
        polling_dic = {}
        with self.assertRaises(Exception) as err:
            self.assertEqual(False, self.appauth._check(polling_dic, 1))
        self.assertEqual('Login failed: processing status format is other than expected', str(err.exception))

    def test_002__check(self):
        """ test _check_processing_status() """
        polling_dic = {'data': {'attributes': {'verificationStatus': 'foo'}}}
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(False, self.appauth._check(polling_dic, 1))
        self.assertIn('INFO:dkb_robo.authentication:Unknown processing status: foo', lcm.output)

    def test_0003__check(self):
        """ test _check_processing_status() """
        polling_dic = {'data': {'attributes': {'verificationStatus': 'processed'}}}
        self.assertEqual(True, self.appauth._check(polling_dic, 1))

    def test_004__check(self):
        """ test _check_processing_status() """
        polling_dic = {'data': {'attributes': {'verificationStatus': 'canceled'}}}
        with self.assertRaises(Exception) as err:
            self.assertEqual(True, self.appauth._check(polling_dic, 1))
        self.assertEqual('2fa chanceled by user', str(err.exception))

    def test_005__check(self):
        """ test _check_processing_status() """
        polling_dic = {'data': {'attributes': {'verificationStatus': 'processing'}}}
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(False, self.appauth._check(polling_dic, 1))
        self.assertIn('INFO:dkb_robo.authentication:Status: processing. Waiting for confirmation', lcm.output)

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_006__print(self, mock_stdout):
        """ test _print_app_2fa_confirmation()"""
        self.appauth._print(None)
        self.assertIn('check your banking app and confirm login...', mock_stdout.getvalue())

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_007__print(self, mock_stdout):
        """ test _print_app_2fa_confirmation()"""
        self.appauth._print('devicename')
        self.assertIn('check your banking app on "devicename" and confirm login...', mock_stdout.getvalue())

    @patch('time.sleep', return_value=None)
    @patch('dkb_robo.authentication.APPAuthentication._check')
    @patch('dkb_robo.authentication.APPAuthentication._print')
    def test_008_finalize(self, mock_confirm, mock_status, _mock_sleep):
        """ test _mfa_finalize() """
        self.appauth.client = Mock()
        self.appauth.client.headers = {}
        self.appauth.client.get.return_value.status_code = 200
        self.appauth.client.get.return_value.json.side_effect = [{'foo1': 'bar1'}, {'data': {'attributes': {'verificationStatus': 'bump'}}}]
        mock_status.side_effects = [False, True]
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertTrue(self.appauth.finalize('challengeid', {'foo': 'bar'}, 'devicename'))
        self.assertIn("ERROR:dkb_robo.authentication:error parsing polling response: {'foo1': 'bar1'}", lcm.output)
        self.assertTrue(mock_confirm.called)

    @patch('time.sleep', return_value=None)
    @patch('dkb_robo.authentication.APPAuthentication._check')
    @patch('dkb_robo.authentication.APPAuthentication._print')
    def test_009_finalize(self, mock_confirm, mock_status, _mock_sleep):
        """ test _mfa_finalize() """
        self.appauth.client = Mock()
        self.appauth.client.headers = {}
        self.appauth.client.get.return_value.status_code = 400
        self.appauth.client.get.return_value.json.side_effect = [{'foo1': 'bar1'}, {'data': {'attributes': {'verificationStatus': 'bump'}}}]
        mock_status.return_value = False
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertFalse(self.appauth.finalize('challengeid', {'foo': 'bar'}, 'devicename'))
        self.assertIn('ERROR:dkb_robo.authentication:Polling request failed. RC: 400', lcm.output)
        self.assertTrue(mock_confirm.called)

class TestTANAuthentication(unittest.TestCase):
    """ test class """

    @patch("requests.Session")
    def setUp(self, mock_session):
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        self.tanauth = TANAuthentication(client=mock_session)
        # self.maxDiff = None

    @patch('dkb_robo.authentication.TANAuthentication._image')
    def test_001__print(self, mock_show):
        """ test _print()"""
        challenge_dic = {}
        self.assertFalse(self.tanauth._print(challenge_dic))
        self.assertFalse(mock_show.called)

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    @patch('dkb_robo.authentication.TANAuthentication._image')
    @patch('builtins.input')
    def test_002__print(self, mock_input, mock_show, mock_stdout):
        """ test _print()"""
        challenge_dic = {'data': {'attributes': {'chipTan': {'headline': 'headline', 'instructions': ['in1', 'in2', 'in3']}}}}
        mock_input.return_value=1234
        self.assertEqual(1234, self.tanauth._print(challenge_dic))
        self.assertIn("headline\n\n1. in1\n\n2. in2\n\n3. in3\n\n", mock_stdout.getvalue())
        self.assertFalse(mock_show.called)

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    @patch('dkb_robo.authentication.TANAuthentication._image')
    @patch('builtins.input')
    def test_003__print(self, mock_input, mock_show, mock_stdout):
        """ test _print()"""
        challenge_dic = {'data': {'attributes': {'chipTan': {'qrData': 'qrData', 'headline': 'headline', 'instructions': ['in1', 'in2', 'in3']}}}}
        mock_input.return_value=1234
        self.assertEqual(1234, self.tanauth._print(challenge_dic))
        self.assertIn("headline\n\n1. in1\n\n2. in2\n\n3. in3\n\n", mock_stdout.getvalue())
        self.assertTrue(mock_show.called)

    @patch("PIL.Image.open")
    def test_004__print(self, mock_open):
        """ test _print()"""
        self.assertFalse(self.tanauth._image('cXJEYXRh'))
        self.assertTrue(mock_open.called)

    @patch('dkb_robo.authentication.TANAuthentication._print')
    def test_005_finalize(self, mock_ctan):
        """ test finalize()"""
        mock_ctan.return_value = 'ctan'
        self.tanauth.client = Mock()
        self.tanauth.client.headers = {}
        self.tanauth.client.post.return_value.status_code = 200
        self.tanauth.client.post.return_value.json.return_value = {'data': {'attributes': {'verificationStatus': 'authorized'}}}
        challenge_dic = {'foo': 'bar'}
        self.assertTrue(self.tanauth.finalize('challengeid', challenge_dic, 'devicename'))
        self.assertTrue(mock_ctan.called)

    @patch('dkb_robo.authentication.TANAuthentication._print')
    def test_006_finalize(self, mock_ctan):
        """ test finalize()"""
        mock_ctan.return_value = 'ctan'
        self.tanauth.client = Mock()
        self.tanauth.client.headers = {}
        self.tanauth.client.post.return_value.status_code = 200
        self.tanauth.client.post.return_value.json.return_value = {'data': {'attributes': {'verificationStatus': 'foo'}}}
        challenge_dic = {'foo': 'bar'}
        self.assertFalse(self.tanauth.finalize('challengeid', challenge_dic, 'devicename'))
        self.assertTrue(mock_ctan.called)

    @patch('dkb_robo.authentication.TANAuthentication._print')
    def test_007_finalize(self, mock_ctan):
        """ test finalize()"""
        mock_ctan.return_value = 'ctan'
        self.tanauth.client = Mock()
        self.tanauth.client.headers = {}
        self.tanauth.client.post.return_value.status_code = 400
        self.tanauth.client.post.return_value.json.return_value = {'data': {'attributes': {'verificationStatus': 'foo'}}}
        self.tanauth.client.post.return_value.text = 'bump'
        challenge_dic = {'foo': 'bar'}
        with self.assertRaises(Exception) as err:
            self.assertFalse(self.tanauth.finalize('challengeid', challenge_dic, 'devicename'))
        self.assertEqual('Login failed: 2fa failed. RC: 400 text: bump', str(err.exception))
        self.assertTrue(mock_ctan.called)


if __name__ == '__main__':

    unittest.main()
